import sys
import subprocess
import json
from pathlib import Path
from attrs import define, field
import cattr

__SUPPORTED_PLATFORMS__ = ["win32"]

assert sys.platform in __SUPPORTED_PLATFORMS__, f"unsupported platform: {sys.platform}"

_bin_dir = Path(__file__).parent / "_bin"
_table2json = _bin_dir / "table2json.exe"
assert _table2json.is_file(), f"panic: jsonizer is not installed?: {_table2json}"


@define
class Rules:
    keys: list[str] = field(default=list())
    map: dict[str, str] = field(default=dict())


@define
class Column:
    name: str = field(default="")
    comment: str = field(default="")
    data_type: str = field(default="")
    data_type_base: str = field(default="")
    scale: int = field(default=None)
    precission: int = field(default=None)
    is_derived: bool = field(default=False)
    is_derived_from: str = field(default=None)
    is_derived_to: str = field(default=None)
    is_validtime: bool = field(default=False)
    is_volatile: bool = field(default=False)
    is_transactiontime: bool = field(default=False)
    compress: str = field(default="")
    character_set: str = field(default="")
    case_specific: bool = field(default=False)
    primary: bool = field(default=False)
    etl_src_not_available: bool = field(default=False)
    etl_crc_format: str = field(default="")
    etl_src_header: str = field(default="")
    etl_src_size_in_bytes: int = field(default=0)
    rules: Rules = field(default=Rules())
    cdc_src_column: str = field(default="")
    cdc_src_data_type: str = field(default="")
    cdc_src_scale: int = field(default=0)


@define
class Table:
    database: str = field(default="")
    name: str = field(default="")
    comment: str = field(default="")
    set_or_multiset: str = field(default="")
    columns: list[Column] = field(default="")
    pi_name: str = field(default=None)
    pi_is_upi: bool = field(default=False)
    pi_columns: list[str] = field(default=list())
    pi_partitions: str = field(default=None)
    pi_isolated_loading: bool = field(default=False)

    def __getitem__(self, key) -> Column:
        if isinstance(key, int):
            return self.columns[key]
        assert isinstance(key, str), f"key is not a str: {key.__class__.__name__}"
        for c in self.columns:
            if c.name.casefold() == key.casefold():
                return c
        raise KeyError(f"{key}: not found in {self.name}")

    def __setitem__(self, key, val) -> None:
        assert isinstance(
            val, Column
        ), f"expected a column instance, got {val.__class__.__name__}"
        if isinstance(key, int):
            self.columns.insert(key, val)
            return
        self.columns.append(val)

    def ___del_from_pi___(self, name: str):
        pos = None
        for i, c in enumerate(self.pi_columns):
            if c.casefold() == name:
                pos = i
                break
        if pos is not None:
            del self.pi_columns[pos]

    def __delitem__(self, name_or_index: int | str | Column):
        colname = None
        if isinstance(name_or_index, Column):
            colname = name_or_index.name.casefold()
        elif isinstance(name_or_index, str):
            colname = name_or_index.casefold()

        if colname:
            pos = -1
            for i, c in enumerate(self.columns):
                if c.name.casefold() == colname:
                    pos = i
                    break
            if pos == -1:
                raise IndexError(f"column not found in {self.name}: {colname}")

            self.___del_from_pi___(colname)
            del self.columns[pos]
            return

        if isinstance(name_or_index, int):
            self.___del_from_pi___(self.columns[name_or_index].name)
            del self.columns[name_or_index]
            return

        raise ValueError(
            f"expected to get one of int, str, Columns, got: {name_or_index.__class__.__name__}"
        )

    def __len__(self) -> int:
        return len(self.columns)

    def __iter__(self):
        return self.columns.__iter__()

    def __contains__(self, item):
        colname = None
        if isinstance(item, Column):
            colname = item.name
        elif isinstance(item, str):
            colname = item
        assert (
            colname
        ), f"expected to get one of str, Column, got: {item.__class__.__name__}"
        for c in self.columns:
            if c.name.casefold() == colname.casefold():
                return True
        return False

    def __reversed__(self):
        return self.columns.__reversed__()


def from_file(filename: str | Path, timeout_seconds=2) -> list[Table] | Table:
    """returns a list of Table instances parsed from DDL script.

    Args:
        filename (str | Path): path to the script
        timeout_seconds (int, optional): safeguard; the parser can freeze on invalid DDL,
            therefore it will be allowed to run only for limited time.
            After the limite it will be killed (and exception will be raised).
    Raises:
        subprocess.TimeoutExpired: signifies that the DDL is likely invalid.
        subprocess.CalledProcessError: signifies that the DDL is likely invalid.

    Returns:
        list[Table] | Table: list of table instances (if found)
    """
    if isinstance(filename, str):
        filename = Path(filename)
    assert isinstance(
        filename, Path
    ), f"filename: not a path: {filename.__class__.__name__}"
    filename = filename.resolve()
    assert filename.is_file(), f"filename: does not exist {filename}"

    jsonizer = [str(_table2json), "-file", str(filename)]

    # check_output raises CalledProcessError on non-zero exit status as specified in the question's text unlike proc.communicate() method.
    #
    try:
        output_bytes = subprocess.check_output(
            jsonizer, stderr=subprocess.STDOUT, timeout=timeout_seconds
        )
    except subprocess.TimeoutExpired:
        raise subprocess.TimeoutExpired(
            f"timeout: {_table2json.name} did not finish in under {timeout_seconds} seconds: check the file {filename}"
        )
    except subprocess.CalledProcessError:
        raise subprocess.CalledProcessError(
            f"error: {_table2json.name} returned error, check the file {filename}"
        )

    # get the json
    output_string = output_bytes.decode("utf-8")
    # get the table
    tables_list = json.loads(output_string)
    tables = cattr.structure(tables_list, list[Table])

    return tables



from collections import namedtuple
from pathlib import Path


LastStatementLine = namedtuple("LastStatementLine", "line_no,offset")


class PatchError(ValueError):
    """PatchError: chyba patch procesu"""

    def __init__(self, message):
        super().__init__(message)
        self.message = message


def async_patch(directory: Path, limit: int = 50, encoding: str = "utf-8"):
    """provede patch TPT skriptů v kontextu použití asynchronní stage"""
    # sestav seznam souborů
    files = list(directory.rglob("*.tpt"))
    if len(files) == 0:
        raise PatchError(f"Nula souborů na vstupu: directory={str(directory)}")
    if len(files) > limit:
        raise PatchError(
            f"Více souborů než jsme čekali.\n- {limit=}\n- {len(files)=}\n- directory={str(directory)}"
            + "\n\nPoužij parametr --limit, pokud chceš limit zvednout."
        )

    # projdi soubor za souborem, seber si informaci o tom kde jsme měli úspěch a kde problém
    problems = []
    for f in files:
        relpath = f.relative_to(directory)
        try:
            script = f.read_text(encoding=encoding, errors="strict")
        except UnicodeError:
            problems.append((relpath, "UnicodeError, oprav kódování a zkus to znovu"))
            continue

        try:
            new_script = _patch_text(script)
        except PatchError as e:
            problems.append((relpath, f"PatchError: {e.message}"))
            continue

        if new_script == script:
            continue

        try:
            f.write_text(new_script, encoding=encoding, errors="strict")
        except Exception as e:
            problems.append((relpath, f"{e}"))

    if len(problems) > 0:
        for p in problems:
            file, problem = p
            print(f"{str(file)}: {problem}")
    else:
        print("ALL OK!")


def _patch_text(text: str):
    lines = text.split("\n")

    # nejdříve si najdeme potřebné pozice - následující MUSÍM mít k dispozici vždy
    i_apply = tpt_line_with_apply(lines)
    i_insert = tpt_line_with_insert(lines)
    i_operator = tpt_line_with_to_operator(lines)
    last_stmt = tpt_line_with_semicolon_above_operator(lines, i_operator)

    # následující můžu a nemusím mít k dispozici
    io_set_session = tpt_line_with_set_session(lines)
    io_restart_prm = tpt_line_with_restart_param(lines, i_operator)
    io_password = tpt_line_with_password(lines, i_operator)

    # pokud již ve skriptu JE set session validtime, končíme a nic neděláme
    if io_set_session > -1:
        return text

    # jinak musíme provést patch
    out_lines = lines.copy()

    # pokud je na řádku s APPLY apostrof, musíme ho "zasmrtit", ale současně ho musíme dodat za SET SESSION VALIDTIME
    i_apply_apostrof = ""
    if "'" in out_lines[i_apply]:
        if out_lines[i_apply].count("'") > 1:
            raise PatchError(f"na řádce s APPLY je víc apostrofů než jeden: {out_lines[i_apply]}")
        i_apply_apostrof = "\n          '\n"
        out_lines[i_apply] = out_lines[i_apply].replace("'", "")

    # řádek, kde je APPLY, obohatíme o SET SESSION VALIDTIME
    # to musíme udělat vždy - set session ve skriptu není (jinak bychom ho vrátili bez úprav)
    out_lines[i_apply] = (
        out_lines[i_apply]
        + "\n          'SET SESSION VALIDTIME AS OF TIMESTAMP ''%LOAD_DT% 23:59:59'';'"
        + "\n          ,"
        + i_apply_apostrof
    )

    # řádek, kde je insert, musíme zabalit do SELECT * FROM (
    # poslední statement musíme zakončit na ) as T
    # ale jen opkud prní select není SELECT * FROM
    if not tpt_is_first_select_star(lines):
        out_lines[i_insert] = out_lines[i_insert] + "\n          SELECT * FROM ("
        out_lines[last_stmt.line_no] = (
            out_lines[last_stmt.line_no][: last_stmt.offset]
            + "\n          ) as T\n"
            + out_lines[last_stmt.line_no][last_stmt.offset :]
        )

    # pokud ještě nemáme  RestartAtFirstDmlGFroup, doplnit
    if io_restart_prm == -1:
        if io_password == -1:
            raise PatchError("""nenašel jsem řádek, kde je "UserPassword" parametr na úrovni TO OPERATOR""")
        out_lines[io_password] = out_lines[io_password] + "\n             " + """, "RestartAtFirstDMLGroup" = 'Yes'"""

    # doplň prázdný řádek, pokud na konci není
    if len(out_lines[-1]) > 0:
        out_lines.append("")

    # vrať zpátky upravený skript ...
    return "\n".join(out_lines)


def tpt_is_first_select_star(in_lines: list[str]) -> bool:
    for line in in_lines:
        line = line.upper().lstrip(",' ")
        line = "".join(line.split(" "))
        if line.startswith("SELECT"):
            if line.startswith("SELECT*FROM"):
                return True
            return False
    raise PatchError("failed to get the select")


def tpt_line_with_apply(in_lines: list[str]) -> int:
    """vrací pozice řádek, kde je APPLY"""
    lines = []
    for i, line in enumerate(in_lines):
        if line.strip("' ").upper() == "APPLY":
            lines.append(i)

    if len(lines) == 0:
        raise PatchError("Nenašel jsem řádek, kde by byl APPLY statement.")
    if len(lines) > 1:
        raise PatchError(f"Příliš mnoho řádků, kde je APPLY: {lines=}")
    return lines[0]


def tpt_line_with_insert(in_lines: list[str]) -> int:
    """vrací pozice řádek, kde je INSERT"""
    lines = []
    for i, line in enumerate(in_lines):
        line = line.replace("'", "").strip().removeprefix(",").strip().upper()
        if line.startswith("INSERT") and "INTO" in line:
            lines.append(i)

    if len(lines) == 0:
        raise PatchError("Nenašel jsem řádek, kde by byl INSERT statement.")
    if len(lines) > 1:
        raise PatchError(f"Příliš mnoho řádků, kde je insert: {lines=}")
    return lines[0]


def tpt_line_with_to_operator(in_lines: list[str]) -> int:
    """vrací pozice řádek, kde je TO OPERATOR"""
    lines = []
    for i, line in enumerate(in_lines):
        line = line.upper().strip()
        if line == "TO OPERATOR":
            lines.append(i)

    if len(lines) == 0:
        raise PatchError("Nenašel jsem řádek, kde by byl INSERT statement.")
    if len(lines) > 1:
        raise PatchError(f"Příliš mnoho řádků, kde je insert: {lines=}")
    return lines[0]


def tpt_line_with_semicolon_above_operator(in_lines: list[str], line_with_to_operator: int) -> LastStatementLine:
    """
    funkce vrací tupple LastStatementLine
    - .line_no - číslo řádku, kde končí poslední statement nad TO OPERATOR
    - .offset - offset, na kterém je středník, nebo apostrof ... tj kde je ukončený statement
    """
    last_statement_line = -1
    for i, line in enumerate(in_lines):
        line = line.replace(" ", "")
        if line == ";'" or line == "'":
            if i > last_statement_line and i <= line_with_to_operator:
                last_statement_line = i
        if i > line_with_to_operator:
            break
    if last_statement_line == -1:
        raise PatchError(
            f"""Nenašel jsem řádek, který by byl nad `TO OPERATOR`, a který by končil středníkem nebo apostrofem.
        Číslo řádku : {line_with_to_operator = }"""
        )

    offset_apostrof = in_lines[last_statement_line].rfind(";")
    if offset_apostrof == -1:
        offset_apostrof = in_lines[last_statement_line].rfind("'")
    if offset_apostrof == -1:
        raise PatchError(f"Na řádce {last_statement_line=} jsem nenašel ani středník ani apostrof")

    return LastStatementLine(line_no=last_statement_line, offset=offset_apostrof)


def tpt_line_with_set_session(in_lines: list[str]) -> int:
    """vrací pozice řádek, kde je SET SESSION VALIDTIME, nebo -1"""
    lines = []
    for i, line in enumerate(in_lines):
        line = line.strip().upper()
        if "SET SESSION VALIDTIME" in line:
            lines.append(i)

    if len(lines) > 1:
        raise PatchError(f"Příliš mnoho řádků, kde je SET SESSION VALIDTIME: {lines=}")

    if len(lines) == 0:
        return -1

    return lines[0]


def tpt_line_with_restart_param(in_lines: list[str], to_operator_line: int) -> int:
    """vrací pozice řádek, kde je RestartAtFirstDMLGroup, nebo -1"""
    lines = []
    for i, line in enumerate(in_lines):
        line = line.strip().upper()
        if "RESTARTATFIRSTDMLGROUP" in line and i > to_operator_line:
            lines.append(i)

    if len(lines) > 1:
        raise PatchError(f"Příliš mnoho řádků, kde je RestartAtFirstDMLGroup: {lines=}")

    if len(lines) == 0:
        return -1

    return lines[0]


def tpt_line_with_password(in_lines: list[str], to_operator_line: int) -> int:
    """vrací pozice řádek, kde je RestartAtFirstDMLGroup, nebo -1"""
    lines = []
    for i, line in enumerate(in_lines):
        line = line.strip().upper()
        if "USERPASSWORD" in line and i > to_operator_line:
            lines.append(i)

    if len(lines) > 1:
        raise PatchError(f"Příliš mnoho řádků, kde je UserPassword: {lines=}")

    if len(lines) == 0:
        return -1

    return lines[0]



import re
from typing import Optional
from collections import namedtuple
from attrs import define, field

BACKTICKS = "```"

KeyAndVal = namedtuple("KeyAndVal", "key val")


class ParsingError(ValueError):
    def __init__(self, message):
        super().__init__(message)


@define
class ColumnMapping:
    column: str
    mapped_to: str = field(default="")
    transformation_rule: str = field(default="")
    comment: str = field(default="")


@define
class Mapping:
    model: str
    data_source: str
    table: str
    name: str

    mapped_to: list[str] = field(factory=list)
    comment: Optional[str] = field(default=None)

    etl_historization: Optional[str] = field(default=None)
    etl_matching_rules: list[str] = field(factory=list)
    etl_main_source: Optional[str] = field(default=None)
    etl_referenced_sources: Optional[str] = field(default=None)
    etl_filter_criterion: Optional[str] = field(default=None)
    etl_active_flag: Optional[str] = field(default=None)
    etl_groupping: Optional[str] = field(default=None)
    etl_calendar: Optional[str] = field(default=None)

    etl_before_insert: Optional[str] = field(default=None)
    etl_after_insert: Optional[str] = field(default=None)

    column_mappings: list[ColumnMapping] = field(factory=list)
    attached_rules: list[str] = field(default=list)
    etl_prep_steps: list[str] = field(factory=list)
    etl_post_steps: list[str] = field(factory=list)

    def __getitem__(self, key):
        col_name = key.lower()
        for cm in self.column_mappings:
            if col_name == cm.column.lower():
                return cm
        raise KeyError(f"column is not mapped: {key}")


def _split(line: str) -> list[str]:
    return re.split(r"[, \n\t]+", line)


def filter_keep_code(lines: list[str]) -> list[str]:
    """
    filter_keep_code: only keep those lines of the text, that are enclosed by backticks

    Args:
        lines (list[str]): lines of text

    Returns:
        list[str]: filtered lines of text

    Throws:
        ParsingException if a code block is not finished.
    """
    in_code = False
    in_skipped_code = False
    result = []
    open_line, at_line = 0, 0

    for line in lines:
        at_line += 1
        stripped_line = line.strip()

        if stripped_line.startswith("# Parse errors"):
            break

        if line.strip() == BACKTICKS:
            if in_skipped_code:
                in_skipped_code = False
            else:
                in_code = not in_code
            if in_code:
                open_line = at_line
            continue

        if stripped_line.startswith(BACKTICKS):
            in_skipped_code = True
            continue

        if not in_code:
            continue
        if stripped_line.startswith("#"):
            continue

        result.append(line)
    if in_code:
        raise ParsingError(f"Problem at line {open_line}: code block is not closed")
    return result


def _is_key_val_property(line: str, throws_on_fail=False) -> Optional[KeyAndVal]:
    line = line.strip()

    try:
        position = line.index("=")
    except ValueError:
        if throws_on_fail:
            raise ParsingError(f"input line is not a key value property: {line}")
        return None

    k = line[:position].upper().strip()
    v = line[(position + 1) :].strip()
    return KeyAndVal(key=k, val=v)


def _is_in_square_brackets(line: str) -> Optional[KeyAndVal]:
    """
    returns the tuple KeyAndVal, if the line is in square brackets.
    throws ParsingError if the tuple is not separated by equal sign

    Args:
        line (str): _description_

    Returns:
        Optional[KeyAndVal]: _description_
    """
    line = line.strip()
    if not (line.startswith("[") and line.endswith("]")):
        return None
    line = line[1:-1]  # strip brackets
    return _is_key_val_property(line, throws_on_fail=True)


def from_markdown(text: str, filter=filter_keep_code) -> Mapping:
    """
    Parses a markdown file and returns mapping.

    Args:
        text (str): _description_
        filter (_type_, optional): _description_. Defaults to filter_keep_code.
    """
    lines = text.split("\n")
    in_multiline = False
    colmap: Optional[ColumnMapping] = None
    line_buffer = []
    property = ""

    ret_map = Mapping(model="", data_source="", table="", name="")
    lines = filter_keep_code(lines)
    for line in lines:
        stripped_line = line.strip()

        if not in_multiline:
            # skip empty lines and comments
            if stripped_line == "":
                continue

            # accept change of context
            ctx_change = _is_in_square_brackets(stripped_line)
            if ctx_change:
                match ctx_change.key:
                    case "MODEL":
                        ret_map.model = ctx_change.val
                    case "TABLE":
                        ret_map.table = ctx_change.val
                    case "DATA_SOURCE":
                        ret_map.data_source = ctx_change.val
                    case "MAPPING":
                        ret_map.name = ctx_change.val
                    case "COLUMN":
                        colmap = ColumnMapping(column=ctx_change.val)
                        ret_map.column_mappings.append(colmap)
                    case "BUSINESS_RULE":
                        ret_map.attached_rules.append(ctx_change.val)
                continue

            # should be a key/val property then ...
            kv = _is_key_val_property(line)
            if not kv:
                raise ParsingError(f"line should be key/val property: {line}")

            # might be a multiline property ....
            # in that case, keep everything after the arrow
            property = kv.key
            skv: str = kv.val.strip()
            if skv.startswith("->"):
                skv = skv.removeprefix("->").strip()
                if skv != "":
                    line_buffer.append(skv)
                in_multiline = True
                continue
        else:
            if stripped_line != "<-":
                line_buffer.append(line)
                continue
            else:
                in_multiline = False

        if property == "":
            raise ParsingError(f"internal error: property is not set: line={line}")

        # store what you have
        if len(line_buffer) > 0:
            stored_value = "\n".join(line_buffer)
        else:
            stored_value = kv.val.lstrip()

        # if column_mapping exists, we are modifying the column mapping
        if colmap and colmap.column != "":
            match property:
                case "MAPPED_TO":
                    colmap.mapped_to = stored_value
                case ("TO2_PDM_EDW.ETL_TRANSFORMATION_RULE" | "TO2_PDM_EDW.ETL_TRANSFORMATION_RULES"):
                    colmap.transformation_rule = stored_value
                case "COMMENT":
                    colmap.comment = stored_value
                case _:
                    raise ParsingError(f"unknown property for column mapping {colmap.column}: {property}")

            line_buffer = []
            continue

        # otherwise it is a table property
        match property:
            case "MAPPED_TO":
                ret_map.mapped_to = sorted(_split(stored_value))

            case "COMMENT":
                ret_map.comment = stored_value
            case "TO2_PDM_EDW.ETL_HISTORIZATION":
                ret_map.etl_historization = stored_value
            case "TO2_PDM_EDW.ETL_MATCHING_RULES":
                ret_map.etl_matching_rules = sorted(_split(stored_value))
            case "TO2_PDM_EDW.ETL_MAIN_SOURCE":
                ret_map.etl_main_source = stored_value
            case "TO2_PDM_EDW.ETL_REFERENCED_SOURCES":
                ret_map.etl_referenced_sources = stored_value
            case "TO2_PDM_EDW.ETL_FILTER_CRITERION":
                ret_map.etl_filter_criterion = stored_value
            case "TO2_PDM_EDW.ETL_TRANSFORMATION_ALGORITHM":
                ret_map.etl_groupping = stored_value
            case "TO2_PDM_EDW.ETL_TRANSFORMATION_RULES":
                ret_map.etl_active_flag = stored_value
            case "TO2_PDM_EDW.ETL_CALENDAR":
                ret_map.etl_calendar = stored_value
            case "TO2_PDM_EDW.ETL_PREP_STEP":
                ret_map.etl_prep_steps.append(stored_value)
            case "TO2_PDM_EDW.ETL_POST_STEP":
                ret_map.etl_prep_steps.append(stored_value)
        line_buffer = []

    return ret_map



from pathlib import Path as _Path
import fire as _fire
from rich import traceback
from genpypress import app_cc as _app_cc
from genpypress import app_patch_to_validtime as _app_patch_to_validtime

traceback.install(show_locals=False, max_frames=1)


def apatch(directory: str, limit: int = 50, encoding: str = "utf-8"):
    """apatch: patch TPT skriptů pro async stage

    Args:
        directory (str): adresář, kde jsou TPT skripty
        limit (int): kolik maximálně souborů upravit
        encoding (str): jak jsou soubory nakódované
    """
    d = _Path(directory)
    if not d.is_dir():
        print(f"toto není adresář: {directory}")
        exit(1)
    _app_patch_to_validtime.async_patch(d, limit, encoding)


def cc(
    directory: str,
    scenario: str = "drop",
    input_encoding: str = "utf-8",
    output_encoding: str = "utf-8",
    max_files: int = 20,
):
    """cc: conditional create

    Args:
        directory (str): directory where to do the work
        scenario (str): ["drop", "create", "cleanup", "drop-only"]
        input_encoding (str): Defaults to "utf-8".
        output_encoding (str): Defaults to "utf-8".
    """
    _app_cc.conditional_create(directory, scenario, input_encoding, output_encoding, max_files)


def _main():
    _fire.Fire()


if __name__ == "__main__":
    _main()



from pathlib import Path
import re


class ContentError(ValueError):
    """ContentError is raised by the app."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


_VALID_SCENARIOS = ["drop", "create", "cleanup", "drop-only"]

# regular expressions building blocks
_WS = r"\s+"
_MAYBE_WS = r"\s*"
_CREATE = "create"
_SET = "set +|multiset +"
_TABLE = "table"
_IDENTIFIER = "[a-z0-9_.]+"
_AS = "as"
_PATTERN = (
    _CREATE
    + _WS
    + "("
    + _SET
    + ")?"
    + _MAYBE_WS
    + _TABLE
    + _WS
    + "(?P<IDENT>"
    + _IDENTIFIER
    + ")"
)
_PATTERN2 = (
    _CREATE
    + _WS
    + _TABLE
    + _WS
    + "(?P<IDENT>"
    + _IDENTIFIER
    + ")"
    + _AS
    + _WS
    + _IDENTIFIER
)

# regular expressions used to find what is needed
_re_create = re.compile(_PATTERN, re.I)
_re_create2 = re.compile(_PATTERN2, re.I)
_re_condition = re.compile(r"\/\*--conditional_create--\*\/", re.I)


# the template inected into SQL files
_TEMPLATE_CREATE = """-----------------------------------------------------------------------
/*--conditional_create--*/
select count(*) from dbc.tablesV 
where tablename='<table>' and databasename=( select database )
having count(*) > 0;

.IF ERRORCODE <> 0 THEN .QUIT 10;
.IF ACTIVITYCOUNT = 0 THEN .GOTO TABLE_DOES_NOT_EXIST;
.REMARK 'tabulka EXISTUJE'
.GOTO KONEC

.LABEL TABLE_DOES_NOT_EXIST
-----------------------------------------------------------------------
"""

_TEMPLATE_DROP = """-----------------------------------------------------------------------
/*--conditional_create--*/
select count(*) from dbc.tablesV 
where tablename='<table>' and databasename=( select database )
having count(*) > 0;

.IF ERRORCODE <> 0 THEN .QUIT 10;
.IF ACTIVITYCOUNT = 0 THEN .GOTO TABLE_DOES_NOT_EXIST;
.REMARK 'tabulka EXISTUJE'
drop table <table>;

.LABEL TABLE_DOES_NOT_EXIST
-----------------------------------------------------------------------
"""


def _load_script(file: Path, input_encoding: str):
    """loads the script, and filters out the header and footer of the conditional create.

    Args:
        file (Path): path to the file
        input_encoding (str): encoding used to load the file

    Returns:
        _type_: _description_
    """
    outLines = []
    with open(file, "r", encoding=input_encoding) as f:
        script = f.read()

        # pokud neobsahuje ".LABEL TABLE_DOES_NOT_EXIST", není tam co řešit
        if ".LABEL TABLE_DOES_NOT_EXIST" not in script:
            return script

        # jinak ale musíme filtrovat
        lines = script.splitlines()
        skipThis = True
        readNext = False

        for line in lines:
            if line == ".LABEL TABLE_DOES_NOT_EXIST":
                readNext = True
                continue
            if readNext:
                skipThis = False
                readNext = False
                continue

            if line == ".LABEL KONEC":
                break

            if skipThis:
                continue

            outLines.append(line)

    return "\n".join(outLines)


def _is_create_script(insides: str, file: str):
    if m := _re_create.search(insides):
        parts = m.group("IDENT").split(".")
        if len(parts) == 1:
            return True, parts[0]
        return True, parts[1]
    elif m := _re_create2.search(insides):
        parts = m.group("IDENT").split(".")
        if len(parts) == 1:
            return True, parts[0]
        return True, parts[1]
    else:
        raise ContentError(f"did not find name of the table: {file}")
        return False, None


def _contains_conditional_create(insides):
    return _re_condition.search(insides)


def _change_to_conditional(
    file, output_encoding: str, input_encoding: str, create_or_drop="create"
):
    try:
        insides = _load_script(file, input_encoding)
    except UnicodeDecodeError as e:
        print("Encoding error")
        print(file)
        raise e

    is_create, creates = _is_create_script(insides, file)

    if not is_create:
        return False

    newScript = ""

    if create_or_drop == "cleanup":
        newScript = insides

    else:
        TEMPLATE = _TEMPLATE_CREATE if create_or_drop == "create" else _TEMPLATE_DROP
        inject = TEMPLATE.replace("<table>", creates)
        if create_or_drop == "drop-only":
            insides = "--only drop the table, do not create\n\n"
        newScript = inject + insides + "\n" + ".LABEL KONEC"

    # save
    try:
        with open(file, "w", encoding=output_encoding) as f:
            f.write(newScript)
    except Exception as e:
        print("chyba, prisel jsi o ", file)
        print(e)


def conditional_create(
    folder,
    scenario="create",
    input_encoding: str = "utf-8",
    output_encoding: str = "utf-8",
    max_files: int = 20,
):
    """
    Scans a folder for files with .sql or .bteq extension, and tries to insert conditional create header and footer in them.
    Raises a ContentError when the number of files is > max_files.

    Args:
        folder (str): input folder
        scenario (str, optional): ["drop", "create", "cleanup", "drop-only"], defaults to "create".
        input_encoding (str, optional): defaults to "utf-8".
        output_encoding (str, optional): defaults to "utf-8".
        max_files (int, optional): defaults to 20.

    Raises:
        ContentError: raised when number of files is too great, or when one of the files can not be patched.
    """
    assert scenario in _VALID_SCENARIOS, f"{scenario}: not one of {_VALID_SCENARIOS}"
    if isinstance(folder, str):
        folder = Path(folder)

    files = list(folder.rglob("*.sql"))
    files.extend(list(folder.rglob("*.bteq")))
    print(f"Files={len(files)} in [{folder}]")

    if len(files) > max_files:
        raise ContentError(f"got {len(files)}, expected {max_files=}")

    for f in files:
        _change_to_conditional(f, output_encoding, input_encoding, scenario)



import os
import sys
import contextlib
import teradatasql

# od verze 3.9 je k dispozici collections.abc.Generator, ve starších verzích typing.Generator
_version = (sys.version_info.major * 100) + sys.version_info.minor
if _version > 309:
    from collections.abc import Generator
else:
    from typing import Generator


def _one_of(keys: list[str], default=None) -> str | None:
    for k in keys:
        try:
            val = os.environ[k]
            return val
        except KeyError:
            continue
    return default


DEFAULT_HOSTNAME = _one_of(["TERADATA_HOSTNAME"], default="edwprod.cz.o2")
DEFAULT_USER = _one_of(["TERADATA_USER", "TO2_DOMAIN_USER"], default="edwprod.cz.o2")
DEFAULT_PASSWORD = _one_of(
    ["TERADATA_PASSWORD", "TO2_DOMAIN_PASSWORD"], default="edwprod.cz.o2"
)


class ParameterError(ValueError):
    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


@contextlib.contextmanager
def connect_teradata(
    hostname: str = DEFAULT_HOSTNAME,
    username: str = DEFAULT_USER,
    password: str = DEFAULT_PASSWORD,
    logmech: str | None = "LDAP",
    tmode: str = "TERA",
) -> Generator[teradatasql.TeradataConnection, None, None]:
    """Simple context manager, which can connect to Teradata using "sensible defaults".

    Args:
        hostname (str): hostname, defaults to DEFAULT_HOSTNAME ("edwprod.cz.o2")
        username (str): defaults to os.environ['TO2_DOMAIN_USER']
        password (str): defaults to os.environ['TO2_DOMAIN_PASSWORD']
        logmech (str, optional): defaults to "LDAP".
        tmode (str): ANSI or TERA; defaults to TERA

    Returns:
        teradatasql.TeradataConnection

    Throws:
        any error supported by the teradatasql module
    """
    if username is None:
        raise ParameterError("username: should not be None")
    if password is None:
        raise ParameterError("password: should not be None")
    with teradatasql.connect(
        host=hostname, user=username, password=password, logmech=logmech
    ) as session:
        yield session
